@@ -200,6 +200,7 @@ typedef enum {
 typedef unsigned long long TranslationTableCharacterAttributes;
 
 typedef struct {
+	TranslationTableOffset offset;
 	TranslationTableOffset next;
 	widechar lookFor;
 	widechar found;
@@ -210,6 +211,7 @@ typedef struct {
 	TranslationTableOffset definitionRule;
 	TranslationTableOffset otherRules;
 	TranslationTableCharacterAttributes attributes;
+	TranslationTableOffset offset;
 	widechar realchar;
 	widechar uppercase;
 	widechar lowercase;
@@ -455,6 +457,7 @@ typedef enum { /* Op codes */
 } TranslationTableOpcode;
 
 typedef struct {
+	TranslationTableOffset offset;
 	TranslationTableOffset charsnext;			/** next chars entry */
 	TranslationTableOffset dotsnext;			/** next dots entry */
 	TranslationTableCharacterAttributes after;  /** character types which must follow */
@@ -495,7 +498,7 @@ typedef struct CharacterClass {
 
 typedef struct RuleName {
 	struct RuleName *next;
-	TranslationTableOffset ruleOffset;
+	TranslationTableOffset rule;
 	widechar length;
 	widechar name[1];
 } RuleName;
