@@ -46,15 +46,13 @@ const char version_etc_copyright[] =
 
 #define AUTHORS "Christian Egli"
 
-#define DIRECTION_FORWARD 0
-#define DIRECTION_BACKWARD 1
-#define DIRECTION_BOTH 2
-#define DIRECTION_DEFAULT DIRECTION_FORWARD
-
-#define HYPHENATION_OFF 0
-#define HYPHENATION_ON 1
-#define HYPHENATION_BRAILLE 2
-#define HYPHENATION_DEFAULT HYPHENATION_OFF
+#define MODE_TRANSLATION_FORWARD 0
+#define MODE_TRANSLATION_BACKWARD 1
+#define MODE_TRANSLATION_BOTH_DIRECTIONS 2
+#define MODE_HYPHENATION 3
+#define MODE_HYPHENATION_BRAILLE 4
+#define MODE_DISPLAY 5
+#define MODE_DEFAULT MODE_TRANSLATION_FORWARD
 
 static void
 print_help(void) {
@@ -319,12 +317,11 @@ read_table(yaml_event_t *start_event, yaml_parser_t *parser, const char *display
 }
 
 static void
-read_flags(yaml_parser_t *parser, int *direction, int *hyphenation) {
+read_flags(yaml_parser_t *parser, int *testmode) {
 	yaml_event_t event;
 	int parse_error = 1;
 
-	*direction = DIRECTION_DEFAULT;
-	*hyphenation = HYPHENATION_DEFAULT;
+	*testmode = MODE_DEFAULT;
 
 	if (!yaml_parser_parse(parser, &event) || (event.type != YAML_MAPPING_START_EVENT))
 		yaml_error(YAML_MAPPING_START_EVENT, &event);
@@ -338,16 +335,18 @@ read_flags(yaml_parser_t *parser, int *direction, int *hyphenation) {
 			if (!yaml_parser_parse(parser, &event) || (event.type != YAML_SCALAR_EVENT))
 				yaml_error(YAML_SCALAR_EVENT, &event);
 			if (!strcmp((const char *)event.data.scalar.value, "forward")) {
-				*direction = DIRECTION_FORWARD;
+				*testmode = MODE_TRANSLATION_FORWARD;
 			} else if (!strcmp((const char *)event.data.scalar.value, "backward")) {
-				*direction = DIRECTION_BACKWARD;
+				*testmode = MODE_TRANSLATION_BACKWARD;
 			} else if (!strcmp((const char *)event.data.scalar.value, "bothDirections")) {
-				*direction = DIRECTION_BOTH;
+				*testmode = MODE_TRANSLATION_BOTH_DIRECTIONS;
 			} else if (!strcmp((const char *)event.data.scalar.value, "hyphenate")) {
-				*hyphenation = HYPHENATION_ON;
+				*testmode = MODE_HYPHENATION;
 			} else if (!strcmp((const char *)event.data.scalar.value,
 							   "hyphenateBraille")) {
-				*hyphenation = HYPHENATION_BRAILLE;
+				*testmode = MODE_HYPHENATION_BRAILLE;
+			} else if (!strcmp((const char *)event.data.scalar.value, "display")) {
+				*testmode = MODE_DISPLAY;
 			} else {
 				error_at_line(EXIT_FAILURE, 0, file_name, event.start_mark.line + 1,
 						"Testmode '%s' not supported\n", event.data.scalar.value);
@@ -556,8 +555,8 @@ read_typeform_string(yaml_parser_t *parser, formtype *typeform, typeforms kind,
 	typeform_len = strlen((const char *)event.data.scalar.value);
 	if (typeform_len != len)
 		error_at_line(EXIT_FAILURE, 0, file_name, event.start_mark.line + 1,
-				"Too many or too typeforms (%i) for word of length %i\n", typeform_len,
-				len);
+				"Too many or too few typeforms (%i) for word of length %i\n",
+				typeform_len, len);
 	update_typeform((const char *)event.data.scalar.value, typeform, kind);
 	yaml_event_delete(&event);
 }
@@ -609,7 +608,7 @@ read_typeforms(yaml_parser_t *parser, int len) {
 }
 
 static void
-read_options(yaml_parser_t *parser, int direction, int wordLen, int translationLen,
+read_options(yaml_parser_t *parser, int testmode, int wordLen, int translationLen,
 		int *xfail, translationModes *mode, formtype **typeform, int **inPos,
 		int **outPos, int *cursorPos, int *cursorOutPos, int *maxOutputLen,
 		int *realInputLen) {
@@ -635,7 +634,7 @@ read_options(yaml_parser_t *parser, int direction, int wordLen, int translationL
 			yaml_event_delete(&event);
 			*mode = read_mode(parser);
 		} else if (!strcmp(option_name, "typeform")) {
-			if (direction != 0) {
+			if (testmode != MODE_TRANSLATION_FORWARD) {
 				error_at_line(EXIT_FAILURE, 0, file_name, event.start_mark.line + 1,
 						"typeforms only supported with testmode 'forward'\n");
 			}
@@ -648,14 +647,14 @@ read_options(yaml_parser_t *parser, int direction, int wordLen, int translationL
 			yaml_event_delete(&event);
 			*outPos = read_outPos(parser, wordLen, translationLen);
 		} else if (!strcmp(option_name, "cursorPos")) {
-			if (direction == 2) {
+			if (testmode == MODE_TRANSLATION_BOTH_DIRECTIONS) {
 				error_at_line(EXIT_FAILURE, 0, file_name, event.start_mark.line + 1,
 						"cursorPos not supported with testmode 'bothDirections'\n");
 			}
 			yaml_event_delete(&event);
 			read_cursorPos(parser, cursorPos, cursorOutPos, wordLen, translationLen);
 		} else if (!strcmp(option_name, "maxOutputLength")) {
-			if (direction == 2) {
+			if (testmode == MODE_TRANSLATION_BOTH_DIRECTIONS) {
 				error_at_line(EXIT_FAILURE, 0, file_name, event.start_mark.line + 1,
 						"maxOutputLength not supported with testmode 'bothDirections'\n");
 			}
@@ -729,8 +728,7 @@ parsed_strlen(char *s) {
 }
 
 static void
-read_test(yaml_parser_t *parser, char **tables, const char *display_table, int direction,
-		int hyphenation) {
+read_test(yaml_parser_t *parser, char **tables, const char *display_table, int testmode) {
 	yaml_event_t event;
 	char *description = NULL;
 	char *word;
@@ -773,7 +771,7 @@ read_test(yaml_parser_t *parser, char **tables, const char *display_table, int d
 
 	if (event.type == YAML_MAPPING_START_EVENT) {
 		yaml_event_delete(&event);
-		read_options(parser, direction, parsed_strlen(word), parsed_strlen(translation),
+		read_options(parser, testmode, parsed_strlen(word), parsed_strlen(translation),
 				&xfail, &mode, &typeform, &inPos, &outPos, &cursorPos, &cursorOutPos,
 				&maxOutputLen, &realInputLen);
 
@@ -789,10 +787,18 @@ read_test(yaml_parser_t *parser, char **tables, const char *display_table, int d
 	char **table = tables;
 	while (*table) {
 		int r;
-		if (hyphenation == HYPHENATION_ON || hyphenation == HYPHENATION_BRAILLE) {
+		if (testmode == MODE_HYPHENATION || testmode == MODE_HYPHENATION_BRAILLE) {
 			r = check_hyphenation(
-					*table, word, translation, hyphenation == HYPHENATION_BRAILLE);
+					*table, word, translation, testmode == MODE_HYPHENATION_BRAILLE);
+
+		} else if (testmode == MODE_DISPLAY) {
+			r = check_display(*table, word, translation);
 		} else {
+			int direction = DIRECTION_FORWARD;
+			if (testmode == MODE_TRANSLATION_BACKWARD)
+				direction = DIRECTION_BACKWARD;
+			else if (testmode == MODE_TRANSLATION_BOTH_DIRECTIONS)
+				direction = DIRECTION_BOTH;
 			// FIXME: Note that the typeform array was constructed using the
 			// emphasis classes mapping of the last compiled table. This
 			// means that if we are testing multiple tables at the same time
@@ -842,8 +848,8 @@ read_test(yaml_parser_t *parser, char **tables, const char *display_table, int d
 }
 
 static void
-read_tests(yaml_parser_t *parser, char **tables, const char *display_table, int direction,
-		int hyphenation) {
+read_tests(
+		yaml_parser_t *parser, char **tables, const char *display_table, int testmode) {
 	yaml_event_t event;
 	if (!yaml_parser_parse(parser, &event) || (event.type != YAML_SEQUENCE_START_EVENT))
 		yaml_error(YAML_SEQUENCE_START_EVENT, &event);
@@ -860,7 +866,7 @@ read_tests(yaml_parser_t *parser, char **tables, const char *display_table, int
 			yaml_event_delete(&event);
 		} else if (event.type == YAML_SEQUENCE_START_EVENT) {
 			yaml_event_delete(&event);
-			read_test(parser, tables, display_table, direction, hyphenation);
+			read_test(parser, tables, display_table, testmode);
 		} else {
 			error_at_line(EXIT_FAILURE, 0, file_name, event.start_mark.line + 1,
 					"Expected %s or %s (actual %s)", event_names[YAML_SEQUENCE_END_EVENT],
@@ -1037,11 +1043,10 @@ main(int argc, char *argv[]) {
 
 		int haveRunTests = 0;
 		while (1) {
-			int direction = DIRECTION_DEFAULT;
-			int hyphenation = HYPHENATION_DEFAULT;
+			int testmode = MODE_DEFAULT;
 			if (!strcmp((const char *)event.data.scalar.value, "flags")) {
 				yaml_event_delete(&event);
-				read_flags(&parser, &direction, &hyphenation);
+				read_flags(&parser, &testmode);
 
 				if (!yaml_parser_parse(&parser, &event) ||
 						(event.type != YAML_SCALAR_EVENT) ||
@@ -1049,12 +1054,12 @@ main(int argc, char *argv[]) {
 					simple_error("tests expected", &parser, &event);
 				}
 				yaml_event_delete(&event);
-				read_tests(&parser, tables, display_table, direction, hyphenation);
+				read_tests(&parser, tables, display_table, testmode);
 				haveRunTests = 1;
 
 			} else if (!strcmp((const char *)event.data.scalar.value, "tests")) {
 				yaml_event_delete(&event);
-				read_tests(&parser, tables, display_table, direction, hyphenation);
+				read_tests(&parser, tables, display_table, testmode);
 				haveRunTests = 1;
 			} else {
 				if (haveRunTests) {
